import hashlib
from importlib import import_module

from django.conf import settings
from django.contrib.auth import get_user
from django.urls import reverse
from django.utils import timezone
from jwcrypto.jwt import JWT

from django_jwt.server.models import Key
from django_jwt.settings_utils import get_setting, get_domain_from_url
from django_jwt.view_utils import calculate_at_hash


def get_value(user, names):
    obj = user
    for name in names:
        obj = getattr(obj, name, None)
        if callable(obj):
            obj = obj()
    return obj


def get_id_token_claims(web, user, scopes, access_token=None, code=None):
    claims = {item.claim: get_value(user, item.attribute_from_user_model.split('.'))
              for item in web.privateclaimswebpage_set.filter(scope__in=scopes)}
    claims.update({'aud': [web.id], 'azp': web.id})
    if access_token is not None:
        claims['at_hash'] = calculate_at_hash(access_token, hashlib.sha256)
    if code is not None:
        claims['c_hash'] = calculate_at_hash(code, hashlib.sha256)
    if claims.get('sub', None) is None:
        claims['sub'] = user.pk
    return claims


def get_access_token_claims(web, user, scopes, jti=None):
    audience = [item.allowed_to.id for item in web.weballowanceotherweb_set.all()]
    audience.append(get_domain_from_url(get_setting('JWT_OIDC.DISCOVERY_ENDPOINT')) + reverse('oidc_userinfo_endpoint'))
    claims = {'aud': audience, 'scope': ' '.join(scopes), 'sub': user.pk}
    if jti is not None:
        claims['jti'] = jti
    return claims


def get_refresh_token_claims(web, user, scopes, external_session):
    audience = [item.allowed_to.id for item in web.weballowanceotherweb_set.all()]
    audience.append(get_domain_from_url(get_setting('JWT_OIDC.DISCOVERY_ENDPOINT')) + reverse('oidc_userinfo_endpoint'))
    claims = {'aud': audience, 'scope': ' '.join(scopes), 'sub': user.pk, 'jti': str(external_session.refresh_token),
              'sid': external_session.id}
    return claims


def create_jwt_code(request, token_type, claims={}, nonce_required=False):
    now = timezone.now()
    token_type = token_type.upper().replace(' ', '_')
    expiration_time = timezone.timedelta(seconds=get_setting('JWT_OIDC.JWT_%s_EXPIRATION_TIME' % token_type))
    jwk = Key.get_actual_jwk()
    claims.update({'iss': get_domain_from_url(get_setting('JWT_OIDC.DISCOVERY_ENDPOINT')),
                   'iat': int(now.timestamp()),
                   'exp': int((now + expiration_time).timestamp())})
    nonce = request.GET.get('nonce', None)
    if nonce_required and nonce is None:
        return None
    elif nonce is not None:
        claims['nonce'] = nonce
    if 'sub' in claims:
        claims['sub'] = str(claims['sub'])
    jwt = JWT(header={'kid': jwk.kid, 'alg': get_setting('JWT_OIDC.SIGNATURE_ALG'), 'typ': 'JWT'},
              claims=claims)
    jwt.make_signed_token(jwk)
    return jwt.serialize()


def get_user_from_request(request, session_id):
    engine = import_module(settings.SESSION_ENGINE)
    request.session = engine.SessionStore(session_id)
    return get_user(request)



from django.contrib.auth import get_user_model
from rest_framework import serializers

from django_jwt.settings_utils import get_setting


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        exclude = get_setting('JWT_OIDC.USERINFO_SERIALIZER_EXCLUDE')



import json
import logging
import types

import urllib3
from django.contrib.auth.models import AnonymousUser
from django.utils import timezone
from django.utils.deprecation import MiddlewareMixin
from jwcrypto.jwt import JWTExpired

from django_jwt.auth import JWTAuthentication
from django_jwt.openid import OpenId2Info
from django_jwt.settings_utils import get_setting


logger = logging.getLogger(__name__)


class JWTAuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        id_token = request.session.get('id_token', None)
        claims = None
        user = None
        if id_token is not None:
            try:
                user, jwt_id_token = JWTAuthentication.authenticate_credentials(id_token)
                claims = json.loads(jwt_id_token.claims)
            except JWTExpired:
                self.refresh_token(request)
                user = getattr(request, 'user', None)
                claims = getattr(request, 'user_claims', None)
            except Exception as e:
                logger.error(e)
                pass
        request.user = user or AnonymousUser()
        request.user_claims = claims or {}
        request.userinfo = request.session.get('userinfo', None)
        self.add_get_access_token_to_request(request)

    def add_get_access_token_to_request(self, request):
        def get_access_token(inner_request):
            expiration_date = inner_request.session.get('expiration_date', None)
            access_token = inner_request.session.get('access_token', None)
            if expiration_date is not None and expiration_date < timezone.now():
                self.refresh_token(inner_request)
                access_token = inner_request.session.get('access_token', None)
            try:
                JWTAuthentication.validate_jwt(access_token)
            except ValueError:
                return access_token
            except (JWTExpired, JWTAuthentication.JWTException):
                self.refresh_token(inner_request)
                access_token = inner_request.session.get('access_token', None)
            return access_token
        request.get_access_token = types.MethodType(get_access_token, request)

    def refresh_token(self, request):
        refresh_token = request.session.get('refresh_token', None)
        if refresh_token is None and OpenId2Info().token_endpoint is not None:
            return None
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': request.session.get('refresh_token', None),
            'client_id': get_setting('JWT_OIDC.CLIENT_ID'),
            'client_secret': get_setting('JWT_OIDC.CLIENT_SECRET'),
            'scope': get_setting('JWT_OIDC.SCOPE'),
        }
        headers = {
            'Content-Type': 'application/json',
        }
        http = urllib3.PoolManager()
        r = http.request('POST', OpenId2Info().token_endpoint, body=json.dumps(data), headers=headers)
        if r.status != 200:
            request.session.flush()
            return None
        response_data = json.loads(r.data.decode('utf-8'))
        if response_data.get('scope', None) and \
                sorted(response_data['scope']) != sorted(get_setting('JWT_OIDC.SCOPE')):
            logger.error('Server returned different scope from configured')
            request.session.flush()
            return None
        refresh_token = response_data.get('refresh_token', None)
        if refresh_token is not None:
            request.session['refresh_token'] = refresh_token
        expires_in = response_data.get('expires_in', None)
        if expires_in is not None:
            request.session['expiration_date'] = timezone.now() + timezone.timedelta(seconds=expires_in)
        id_token = response_data.get('id_token', None)
        if id_token is not None:
            try:
                request.user, jwt_id_token = JWTAuthentication.authenticate_credentials(id_token)
                request.session['id_token'] = id_token
                request.user_claims = json.loads(jwt_id_token.claims)
            except Exception as e:
                logger.warning(e)
                request.user = AnonymousUser()
                request.session.flush()
        access_token = response_data.get('access_token', None)
        if access_token is not None:
            request.session['access_token'] = access_token



from django.conf import settings


class NoDefault:
    pass


JWT_DEFAULT_SETTINGS = {
    # 'JWT_OIDC.DISCOVERY_ENDPOINT': Required
    # 'JWT_OIDC.CLIENT_ID': Required
    # 'JWT_OIDC.TYPE': Required ['fake', 'client', 'provider']
    # 'JWT_OIDC.CLIENT_SECRET': Required if code in RESPONSE_TYPE
    # 'JWT_OIDC.RESPONSE_TYPE': Required
    'JWT_OIDC.ID_TOKEN_RENAME_ATTRIBUTES': {},
    'JWT_OIDC.USER_DEFAULT_ATTRIBUTES': {},
    'JWT_OIDC.CREATE_USER': False,
    'JWT_OIDC.REQUEST_RESPONSE_TYPE': '#',
    'JWT_OIDC.PKCE_EXTENSION': False,
    'JWT_OIDC.CODE_CHALLENGE_METHOD': 'S256',     # ['plain', 'S256'] plain is not supported
    'JWT_OIDC.SCOPE': 'openid',
    'JWT_OIDC.IDENTIFICATION_CLAIM': 'sub',
    'JWT_OIDC.JWK_EXPIRATION_TIME': 3600,
    'JWT_OIDC.JWT_ID_TOKEN_EXPIRATION_TIME': 2700,
    'JWT_OIDC.JWT_ACCESS_TOKEN_EXPIRATION_TIME': 600,
    'JWT_OIDC.JWT_REFRESH_TOKEN_EXPIRATION_TIME': 3600,
    'JWT_OIDC.SIGNATURE_ALG': 'ES512',
    'JWT_OIDC.MAX_REFRESH': 10,
    'JWT_OIDC.USERINFO_SERIALIZER': 'django_jwt.server.serializers.UserSerializer',
    'JWT_OIDC.USERINFO_SERIALIZER_EXCLUDE': ['password'],
    'JWT_OIDC.CLIENT_DISPLAY': None,
    'JWT_OIDC.CLIENT_PROMPT': None,
    'JWT_OIDC.CLIENT_MAX_AGE': None,
    'JWT_OIDC.CLIENT_UI_LOCALES': None,
    'JWT_OIDC.CLIENT_CLAIMS_LOCALES': None,
    'JWT_OIDC.CLIENT_ID_TOKEN_HINT': None,
    'JWT_OIDC.CLIENT_LOGIN_HINT': None,
    'JWT_OIDC.CLIENT_ACR_VALUES': None,
}


JWT_DEFAULT_CONVERTERS = {

}


def get_setting(names):
    default = JWT_DEFAULT_SETTINGS.get(names, NoDefault)
    value = settings
    for name in names.split('.'):
        if isinstance(value, dict):
            value = value.get(name, default)
        else:
            value = getattr(value, name, default)
    if value == NoDefault:
        raise Exception('Setting %s is required' % names)
    converter = JWT_DEFAULT_CONVERTERS.get(names, lambda x: x)
    value = converter(value)
    return value


def get_domain_from_url(url):
    return '/'.join(url.split('/')[0:3])


def get_max_time_token():
    return max([get_setting('JWT_OIDC.JWT_ID_TOKEN_EXPIRATION_TIME'),
                get_setting('JWT_OIDC.JWT_ACCESS_TOKEN_EXPIRATION_TIME'),
                get_setting('JWT_OIDC.JWT_REFRESH_TOKEN_EXPIRATION_TIME')])



from django import forms

from django_jwt.server.models import WebPage


class WebPageAdminForm(forms.ModelForm):
    class Meta:
        model = WebPage
        exclude = ()
        labels = {
            'logout_all': 'Log out from main page',
            'response_type': 'Log in response type'
        }
        help_texts = {
            'host': 'Web page url without the last "/". Example: https://localhost:8000',
            'response_type': 'Hash response is for frontend auth and Params for backend',
        }



import json
import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.http import urlencode
from django.utils.module_loading import import_string
from django.views import View
from django.views.generic import TemplateView
from jwcrypto.jwt import JWTExpired
from rest_framework import viewsets, mixins
from rest_framework.permissions import IsAuthenticated

from django_jwt.auth import JWTAuthentication
from django_jwt.rest_framework import JWTTokenAuthentication

from django_jwt.settings_utils import get_setting, get_domain_from_url

from django_jwt.server.models import WebPage, UserWebPagePermission, UserExternalSession, Key, PrivateClaimsWebPage, \
    NonceUsed
from django_jwt.server.views_utils import create_jwt_code, get_value, get_id_token_claims, get_access_token_claims, \
    get_refresh_token_claims, get_user_from_request

logger = logging.getLogger(__name__)


class OpenIdConfiguration(View):
    def get(self, request):
        alg_types = [alg_type + alg_size for alg_type in ['RS', 'PS', 'ES'] for alg_size in ['256', '284', '512']]
        configuration = {
            'issuer': request.build_absolute_uri('/'),
            'authorization_endpoint': request.build_absolute_uri(reverse('oidc_authorization_endpoint')),
            'token_endpoint': request.build_absolute_uri(reverse('oidc_token_endpoint')),
            'userinfo_endpoint': request.build_absolute_uri(reverse('oidc_userinfo_endpoint')),
            'jwks_uri': request.build_absolute_uri(reverse('oidc_jwks_endpoint')),
            'end_session_endpoint': request.build_absolute_uri(reverse('oidc_end_session_endpoint')),
            'response_types_supported': ['id_token', 'token', 'code'],
            'subject_types_supported': ['public'],
            'id_token_signing_alg_values_supported': alg_types,
            'claims_supported': ['sub', 'iss', 'aud', 'exp', 'iat', 'jti', 'scope', 'azp', 'sid', 'at_hash', 'c_hash',
                                 'nonce'],
            'require_request_uri_registration': True,
        }
        return JsonResponse(configuration)


class AuthorizationView(LoginRequiredMixin, TemplateView):
    template_name = 'django_jwt/oidc_confirmation.html'

    def get_login_url(self):
        connection = self.request.GET.get('connection', '')
        return super().get_login_url() + (('-' + connection) if connection else '')

    def dispatch(self, request, *args, **kwargs):
        try:
            kwargs['web'] = WebPage.objects.prefetch_related('privateclaimswebpage_set') \
                .get(id=self.request.GET.get('client_id'))
        except WebPage.DoesNotExist:
            raise PermissionDenied()
        redirect_url = self.request.GET.get('redirect_uri', None)
        if redirect_url is None or not redirect_url.startswith(kwargs['web'].host + '/'):
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)

    def prepare_scopes_for_template(self, scopes):
        if 'openid' in scopes:
            scopes.remove('openid')

    def required_to_log_in(self):
        max_age = self.request.GET.get('max_age', None)
        login_date = self.request.session.get('login_date', None)
        if max_age is None or login_date is None:
            return False
        max_login_age = timezone.now() - timezone.timedelta(seconds=int(max_age))
        return login_date < max_login_age

    def check_prompt_and_max_age(self):
        prompt = self.request.GET.get('prompt', 'none')
        if (prompt == 'login' and self.request.session.get('oidc_loging_again', None) is None) or \
                self.required_to_log_in():
            self.request.session['oidc_loging_again'] = True
            return redirect(get_setting('LOGIN_URL') + '?' + urlencode({'next': self.request.get_full_path()}))
        if prompt == 'select_account' and self.request.session.get('oidc_select_user', None) is None:
            self.request.session['oidc_select_user'] = True
            return redirect(get_setting('SELECT_USER_URL') + '?' +
                            urlencode({'next': self.request.get_full_path()}))
        if self.request.session.get('oidc_select_user', None) is not None:
            del self.request.session['oidc_select_user']
        if self.request.session.get('oidc_loging_again', None) is not None:
            del self.request.session['oidc_loging_again']
        return None

    def get(self, request, *args, **kwargs):
        response = self.check_prompt_and_max_age()
        if response is not None:
            return response
        web = kwargs.get('web')
        self.check_user_permissions(web)
        scopes = self.request.GET.get('scope').split(' ')
        if 'openid' not in scopes:
            return SuspiciousOperation("No openid scope found; see documentation for correct parameters")
        not_accepted_scopes, accepted_scopes = UserWebPagePermission.get_not_accepted_and_accepted_scopes(
            user=request.user, web=web, scopes=scopes)
        if self.request.GET.get('prompt', 'none') == 'consent' or len(not_accepted_scopes) != 0:
            self.prepare_scopes_for_template(not_accepted_scopes)
            self.prepare_scopes_for_template(accepted_scopes)
            return render(request, self.template_name, context={'web': web, 'not_accepted_scopes': not_accepted_scopes,
                                                                'accepted_scopes': accepted_scopes})
        return self.success_response(web)

    def success_response(self, web):
        nonce = self.request.GET.get('nonce', None)
        if nonce is not None and NonceUsed.is_used(nonce):
            raise PermissionDenied()
        external_session = UserExternalSession.objects \
            .get_or_create(web=web, session_id=self.request.session.session_key)[0]
        scopes = self.request.GET.get('scope').split(' ')
        tokens = self.request.GET.get('response_type', '').split(' ')
        if len(tokens) == 0:
            raise SuspiciousOperation("Invalid request; see documentation for correct parameters")
        response_type = '#'
        response_parameters = {'state': self.request.GET.get('state')}
        if 'token' in tokens:
            claims = get_access_token_claims(web=web, user=self.request.user, scopes=scopes,
                                             jti=external_session.get_access_token_id())
            response_parameters['access_token'] = create_jwt_code(request=self.request, claims=claims,
                                                                  nonce_required=True, token_type='access_token')
            if response_parameters['access_token'] is None:
                raise SuspiciousOperation("Invalid request; nonce missing! see documentation for correct parameters")
            external_session.access_token_sent = True
        if 'code' in tokens:
            nonce = self.request.GET.get('nonce', None)
            if nonce is not None:
                external_session.nonce = nonce
            response_parameters['code'] = external_session.get_authorization_code()
            response_type = '?'
            external_session.scopes = ' '.join(scopes)
        if 'id_token' in tokens:
            claims = get_id_token_claims(web=web, user=self.request.user, scopes=scopes,
                                         access_token=response_parameters.get('access_token', None),
                                         code=response_parameters.get('code', None))
            response_parameters['id_token'] = create_jwt_code(request=self.request, claims=claims, nonce_required=True,
                                                              token_type='id_token')
            if response_parameters['id_token'] is None:
                raise SuspiciousOperation("Invalid request; nonce missing! see documentation for correct parameters")
            external_session.id_token_sent = True
        redirect_uri = self.request.GET.get('redirect_uri')
        external_session.save()
        if response_type in redirect_uri:
            redirect_uri += '&' + urlencode(response_parameters)
        else:
            redirect_uri += response_type + urlencode(response_parameters)
        return redirect(redirect_uri)

    def post(self, request, *args, **kwargs):
        web = kwargs.get('web')
        if request.POST.get('confirmation', 'false') == 'true':
            scopes = self.request.GET.get('scope').split(' ')
            UserWebPagePermission.update_accepted_scopes(user=request.user, web=web, scopes=scopes)
            return redirect(request.get_full_path())
        return render(request, self.template_name, context={'web': web, 'denied': True})

    def check_user_permissions(self, web: WebPage):
        for restrict in web.restrictuserstoweb_set.all():
            if str(get_value(self.request.user, restrict.attribute_from_user_model.split('.'))) == restrict.value:
                raise PermissionDenied('You don\'t have permission to log in into this web page.')


class TokenView(View):
    def check_client_secret(self):
        try:
            web = WebPage.objects.prefetch_related('privateclaimswebpage_set') \
                .get(id=self.request.POST.get('client_id', None))
        except WebPage.DoesNotExist:
            raise PermissionDenied()
        client_secret = self.request.POST.get('client_secret', None)
        if client_secret is not None and web.client_secret != client_secret:
            raise PermissionDenied()
        return web

    def authorization_code_response(self, web):
        code_verifier = self.request.POST.get('code_verifier', None)
        if self.request.POST.get('client_secret', None) is None and code_verifier is None:
            raise PermissionDenied()
        external_session = UserExternalSession.check_authorization_code(self.request.POST.get('code', None))
        if external_session is None:
            raise PermissionDenied()
        if code_verifier is not None and external_session.check_code_challenge(code_verifier):
            raise PermissionDenied()
        user = get_user_from_request(request=self.request, session_id=external_session.session_id)
        scopes = external_session.scopes.split(' ')
        external_session.access_token_sent = external_session.id_token_sent = False
        response_object = self.create_tokens(web=web, user=user, external_session=external_session, scopes=scopes,
                                             access_token=not external_session.access_token_sent,
                                             id_token=not external_session.id_token_sent)
        response_object.update({'token_type': 'Bearer'})
        return JsonResponse(response_object)

    def add_optional_nonce(self, claims, external_session):
        if external_session.nonce != '':
            claims['nonce'] = external_session.nonce

    def create_tokens(self, web, user, external_session, scopes, access_token=True, id_token=True):
        response_object = {}
        if access_token:
            claims = get_access_token_claims(web=web, user=user, scopes=scopes,
                                             jti=external_session.get_access_token_id())
            self.add_optional_nonce(claims, external_session)
            response_object['access_token'] = create_jwt_code(request=self.request, claims=claims,
                                                              token_type='access_token')
            external_session.access_token_sent = True
        if id_token:
            claims = get_id_token_claims(web=web, user=user, scopes=scopes,
                                         access_token=response_object.get('access_token', None),
                                         code=self.request.POST.get('code', None))
            self.add_optional_nonce(claims, external_session)
            response_object['id_token'] = create_jwt_code(request=self.request, claims=claims, token_type='id_token')
            external_session.id_token_sent = True
        external_session.nonce = ''
        external_session.save()
        claims = get_refresh_token_claims(web=web, user=user, scopes=scopes, external_session=external_session)
        response_object['refresh_token'] = create_jwt_code(request=self.request, claims=claims,
                                                           token_type='refresh_token')
        return response_object

    def refresh_token_response(self, web: WebPage):
        try:
            sub_attr = web.privateclaimswebpage_set.get(claim='sub').attribute_from_user_model.replace('.', '__')
        except PrivateClaimsWebPage.DoesNotExist:
            sub_attr = 'pk'
        try:
            user, claims = JWTAuthentication.validate_authorization_jwt(key=self.request.POST.get('refresh_token',
                                                                                                  None),
                                                                        sub_attr=sub_attr)
        except (JWTExpired, JWTAuthentication.JWTException):
            raise PermissionDenied()
        scopes = claims.get('scope', '').split(' ')
        try:
            external_session = UserExternalSession.objects.get(id=claims.get('sid', None))
        except UserExternalSession.DoesNotExist:
            raise PermissionDenied()
        if str(external_session.refresh_token) != str(claims.get('jti', 0)):
            external_session.delete()
            raise PermissionDenied()
        max_refresh = get_setting('JWT_OIDC.MAX_REFRESH')
        if max_refresh is not None and external_session.refresh_token >= max_refresh:
            raise PermissionDenied()
        external_session.refresh_token += 1
        external_session.save()
        response_object = self.create_tokens(web=web, user=user, external_session=external_session, scopes=scopes)
        return JsonResponse(response_object)

    def post(self, request, *args, **kwargs):
        self.request.POST = json.loads(request.body.decode('utf-8'))
        web = self.check_client_secret()
        grant_type = self.request.POST.get('grant_type', None)
        if grant_type == 'refresh_token':
            return self.refresh_token_response(web)
        elif grant_type == 'authorization_code':
            return self.authorization_code_response(web)
        raise SuspiciousOperation("Invalid request; see documentation for correct parameters")


class JWKsView(View):
    def get(self, request):
        return JsonResponse({'keys': Key.get_jwk_set()})


class LogoutView(View):
    def dispatch(self, request, *args, **kwargs):
        token = request.GET.get('id_token_hint', None)
        redirect_url = self.request.GET.get('post_logout_redirect_uri', '')
        host = get_domain_from_url(redirect_url)
        if token is None:
            raise PermissionDenied()
        try:
            kwargs['web'] = WebPage.objects.get(host=host)
        except WebPage.DoesNotExist:
            raise PermissionDenied()
        try:
            kwargs['user'] = JWTAuthentication.authenticate_credentials(token, client_id=kwargs['web'].id)[0]
        except Exception as e:
            logger.warning(e)
        return super().dispatch(request, *args, **kwargs)

    def get_user(self, token, web: WebPage):
        try:
            jwt = JWTAuthentication.validate_jwt(token)
        except JWTAuthentication.JWTException:
            raise PermissionDenied()
        claims = json.loads(jwt.claims)
        if not self.check_aud(claims, web):
            raise PermissionDenied()
        if not JWTAuthentication.verify_claims(claims, client_id=web.id):
            raise PermissionDenied()
        try:
            att = web.privateclaimswebpage_set.get(claim='sub').attribute_from_user_model.replace('.', '__')
        except PrivateClaimsWebPage.DoesNotExist:
            att = 'pk'
        kwargs = {att: claims.get('sub')}
        User = get_user_model()
        try:
            return User.objects.get(**kwargs)
        except User.DoesNotExist:
            raise PermissionDenied()

    def check_aud(self, claims, web):
        if isinstance(claims['aud'], str):
            claims['aud'] = [claims['aud']]
        for aud in claims['aud']:
            if aud in [web.host, web.id]:
                return True
        return False

    def get(self, request, *args, **kwargs):
        if kwargs['web'].logout_all:
            if request.user.is_authenticated and kwargs.get('user', None) is not None:
                if kwargs['user'].pk != request.user.pk:
                    raise PermissionDenied()
            else:
                UserExternalSession.objects.filter(session_id=request.session.session_key).delete()
            request.session.flush()
        if not kwargs['web'].logout_all:
            UserExternalSession.objects.filter(web=kwargs['web'], session_id=request.session.session_key).delete()
        return redirect(request.GET.get('post_logout_redirect_uri'))


class UserInfoViewSet(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
    authentication_classes = [JWTTokenAuthentication]
    permission_classes = [IsAuthenticated]

    def get_serializer_class(self):
        return import_string(get_setting('JWT_OIDC.USERINFO_SERIALIZER'))

    def get_object(self):
        return self.request.user



import json
import logging

import urllib3
from jwcrypto.jwk import JWKSet

from django_jwt.exceptions import JWTClientException
from django_jwt.patterns import Singleton
from django_jwt.settings_utils import get_setting


# singleton class that saves all information from the OpenId server
class OpenId2Info(metaclass=Singleton):
    def __init_local__(self):
        apps = get_setting('INSTALLED_APPS')
        if 'django_jwt.server' not in apps:
            raise JWTClientException('django_jwt.server not installed for type local')

    def __init_remote__(self):
        url = get_setting('JWT_OIDC.DISCOVERY_ENDPOINT')
        # Getting urls info from OpenId server
        http = urllib3.PoolManager()
        r = http.request('GET', url)
        if r.status != 200:
            error_message = 'OpenID returned error code %s on openid-configuration: %s' % (r.status, url)
            logger = logging.getLogger(__name__)
            logger.critical(error_message)
            raise JWTClientException(error_message)
        data = json.loads(r.data.decode('UTF-8'))
        self.jwks_uri = data.get('jwks_uri', None)
        self.token_endpoint = data.get('token_endpoint', None)
        self.authorization_endpoint = data.get('authorization_endpoint', None)
        self.end_session_endpoint = data.get('end_session_endpoint', None)
        self.userinfo_endpoint = data.get('userinfo_endpoint', None)
        if self.jwks_uri is None or self.authorization_endpoint is None:
            raise JWTClientException('jkws_uri or authorization_endpoint not found in OpenId openid-configuration url')

    def __init__(self):
        self.jwks_uri = None
        self.authorization_endpoint = None
        self.end_session_endpoint = None
        self.jwks = None
        self.userinfo_endpoint = None
        self.client_type = get_setting('JWT_OIDC.TYPE')
        if self.client_type == 'provider':
            self.__init_local__()
        elif self.client_type == 'client':
            self.__init_remote__()
        self.fetch_jwks()

    # Updates all the actual JWKS from the OPenId server
    def fetch_jwks(self):
        logger = logging.getLogger(__name__)
        logger.info('Fetching JWKs')
        if self.client_type == 'client':
            http = urllib3.PoolManager()
            r = http.request('GET', self.jwks_uri)
            if r.status != 200:
                error_message = 'OpenID returned error code %s on jwks_uri: %s' % (r.status, self.jwks_uri)
                logger.critical(error_message)
                raise JWTClientException(error_message)
            jwk_set_json = r.data.decode('UTF-8')
        else:
            from django_jwt.server.models import Key
            jwk_set_json = json.dumps({'keys': Key.get_jwk_set()})
        self.jwks = JWKSet.from_json(jwk_set_json)



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from django_jwt.server import views

urlpatterns = [
    path('jwks', views.JWKsView.as_view(), name="oidc_jwks_endpoint"),
    path('authenticate', views.AuthorizationView.as_view(), name="oidc_authorization_endpoint"),
    path('.well-known/openid-configuration', views.OpenIdConfiguration.as_view(), name='oidc_discovery_endpoint'),
    path('userinfo', views.UserInfoViewSet.as_view({'get': 'retrieve'}), name='oidc_userinfo_endpoint'),
    path('logout', views.LogoutView.as_view(), name='oidc_end_session_endpoint'),
    path('token', csrf_exempt(views.TokenView.as_view()), name='oidc_token_endpoint'),
]



from django.contrib import admin

from django_jwt.server.forms import WebPageAdminForm
from django_jwt.server.models import WebPage, PrivateClaimsWebPage, RestrictUsersToWeb, WebAllowanceOtherWeb


class PrivateClaimsWebPageAdmin(admin.StackedInline):
    verbose_name = 'ID Token extra claim'
    model = PrivateClaimsWebPage
    extra = 1


class RestrictUsersToWebAdmin(admin.StackedInline):
    verbose_name = 'User attribute restricted'
    verbose_name_plural = 'User attributes restricted'
    model = RestrictUsersToWeb
    extra = 1


class WebAllowanceOtherWebAdmin(admin.StackedInline):
    verbose_name = 'Allow 3rd party web'
    model = WebAllowanceOtherWeb
    extra = 1
    fk_name = 'web'


@admin.register(WebPage)
class WebPageFullAdmin(admin.ModelAdmin):
    readonly_fields = ('id', 'client_secret')
    inlines = [PrivateClaimsWebPageAdmin, RestrictUsersToWebAdmin, WebAllowanceOtherWebAdmin]
    form = WebPageAdminForm



import json
import logging

from django.contrib.auth import get_user_model
from jwcrypto.jwt import JWT, JWTMissingKey, JWTExpired

from django_jwt.openid import OpenId2Info
from django_jwt.settings_utils import get_setting, get_domain_from_url


logger = logging.getLogger(__name__)


class JWTAuthentication:
    class JWTException(Exception):
        pass

    @classmethod
    def authenticate_credentials(cls, key, nonce=None, client_id=None):
        try:
            jwt = cls.validate_jwt(key)
        except Exception as e:
            if isinstance(e, JWTExpired):
                raise e
            logger.warning(e)
            raise cls.JWTException('Token is not valid')
        claims = json.loads(jwt.claims)
        if not cls.verify_claims(claims, nonce, client_id):
            raise cls.JWTException('Token is not valid')
        user, created = cls.get_or_create_user(profile=claims)
        return user, jwt

    @classmethod
    def validate_authorization_jwt(cls, key, nonce=None, sub_attr='pk'):
        try:
            jwt = cls.validate_jwt(key)
        except Exception as e:
            if isinstance(e, JWTExpired):
                raise e
            logger.warning(e)
            raise cls.JWTException('Token is not valid')
        claims = json.loads(jwt.claims)
        if not cls.verify_claims(claims, nonce):
            raise cls.JWTException('Token is not valid')
        model = get_user_model()
        user = None
        try:
            user = model.objects.get(**{sub_attr: claims['sub']})
        except model.DoesNotExist:
            pass
        return user, claims

    @classmethod
    def validate_jwt(cls, token, second=False):
        try:
            jwt = JWT(jwt=token, key=OpenId2Info().jwks)
        except JWTMissingKey:
            if second:
                raise cls.JWTException('JWK not found')
            OpenId2Info().fetch_jwks()
            jwt = cls.validate_jwt(token=token, second=True)
        except ValueError as e:
            logger = logging.getLogger(__name__)
            logger.info(e)
            raise e
        return jwt

    @classmethod
    def get_defaults_from_claims(cls, claims):
        model = get_user_model()
        translation = {'sub': 'pk'}
        translation.update(get_setting('JWT_OIDC.ID_TOKEN_RENAME_ATTRIBUTES'))
        model_fields = [field.name for field in model._meta.get_fields()] + ['pk']
        defaults = get_setting('JWT_OIDC.USER_DEFAULT_ATTRIBUTES')
        defaults.update({translation.get(key, key): value for key, value in claims.items()
                         if translation.get(key, key) in model_fields})
        for key, value in defaults.items():
            change_field = getattr(model, 'change_%s' % key, None)
            if change_field is not None:
                defaults[key] = change_field(model, value)
        claim_id = get_setting('JWT_OIDC.IDENTIFICATION_CLAIM')
        user_id = translation.get(claim_id, claim_id)
        if user_id not in defaults:
            return None, False
        kwargs = {user_id: defaults[user_id]}
        del defaults[user_id]
        return defaults, kwargs

    @classmethod
    def get_or_create_user(cls, profile):
        model = get_user_model()
        defaults, kwargs = cls.get_defaults_from_claims(profile)
        if get_setting('JWT_OIDC.CREATE_USER'):
            return model.objects.get_or_create(defaults=defaults, **kwargs)
        try:
            return model.objects.get(**kwargs), False
        except model.DoesNotExist:
            return None, False

    @classmethod
    def verify_claims(cls, claims, nonce=None, client_id=None):
        openid_domain = get_domain_from_url(get_setting('JWT_OIDC.DISCOVERY_ENDPOINT'))
        if client_id is None and 'django_jwt.server' in get_setting('INSTALLED_APPS'):
            if not any([aud.startswith(openid_domain) for aud in claims.get('aud', [])]):
                return False
        else:
            if client_id is None:
                client_id = get_setting('JWT_OIDC.CLIENT_ID')
            if client_id not in claims.get('aud', []):
                logger.info('Client id not in aud')
                return False
        openid_domain = get_domain_from_url(get_setting('JWT_OIDC.DISCOVERY_ENDPOINT'))
        iss_domain = get_domain_from_url(claims.get('iss', ''))
        # openid_domain can't be '' because of apps.py
        if openid_domain != iss_domain:
            logger.info('Issuer %s is not from OPENID2_URL %s' % (claims.get('iss', ''), openid_domain))
            return False
        if nonce is not None and claims.get('nonce') != nonce:
            logger.info('Nonce changed')
            return False
        return True



import base64
import hashlib
import json

from Crypto.PublicKey import RSA, ECC

import uuid

from Crypto.Util.py3compat import tobytes
from django.conf import settings
from django.contrib.sessions.models import Session
from django.db import models
from django.utils import timezone
from jwcrypto.jwk import JWK

from django_jwt.settings_utils import get_setting, get_max_time_token


def generate_key():
    alg = get_setting('JWT_OIDC.SIGNATURE_ALG')
    if alg[:2] in ['RS', 'PS']:
        private_key = RSA.generate(2048)
        return tobytes(private_key.export_key(passphrase=get_setting('SECRET_KEY')))
    # ECC
    curve = alg[2:]
    if curve == '512':
        curve = '521'
    private_key = ECC.generate(curve='p%s' % curve)
    return tobytes(private_key.export_key(format='PEM', passphrase=get_setting('SECRET_KEY'),
                                          protection='PBKDF2WithHMAC-SHA1AndAES128-CBC'))


class Key(models.Model):
    kid = models.CharField(primary_key=True, default=uuid.uuid4, editable=False, max_length=40)
    private_key = models.BinaryField(editable=False, default=generate_key)
    date = models.DateTimeField(default=timezone.now, editable=False)

    def to_jwk(self):
        jwk = JWK()
        jwk.import_from_pem(tobytes(self.private_key), kid=str(self.kid), password=tobytes(get_setting('SECRET_KEY')))
        return jwk

    @classmethod
    def get_actual_jwk(cls):
        now = timezone.now()
        expiration_time = timezone.timedelta(seconds=get_setting('JWT_OIDC.JWK_EXPIRATION_TIME'))
        try:
            key = cls.objects.get(date__gt=(now - expiration_time))
        except cls.DoesNotExist:
            key = cls()
            key.save()
        except cls.MultipleObjectsReturned:
            key = cls.objects.filter(date__gt=(now - expiration_time)).order_by('-date').first()
        return key.to_jwk()

    @classmethod
    def get_jwk_set(cls):
        now = timezone.now()
        key_expiration_time = timezone.timedelta(seconds=get_setting('JWT_OIDC.JWK_EXPIRATION_TIME'))
        max_token_expiration_time = timezone.timedelta(seconds=get_max_time_token() + 120)
        cls.objects.filter(date__lt=(now - key_expiration_time - max_token_expiration_time)).delete()
        keys = cls.objects.all()
        return [key.to_jwk().export_public(as_dict=True) for key in keys]


class WebPage(models.Model):
    RESPONSE_PARAMS = '?'
    RESPONSE_HASH = '#'
    RESPONSE_TYPES = [
        (RESPONSE_HASH, 'Hash'),
        (RESPONSE_PARAMS, 'Params')
    ]

    id = models.CharField(primary_key=True, default=uuid.uuid4, editable=False, max_length=40, verbose_name='Client id')
    client_secret = models.CharField(default=uuid.uuid4, editable=False, max_length=40)
    host = models.CharField(unique=True, max_length=200)
    logout_all = models.BooleanField(default=True)
    name = models.CharField(blank=True, max_length=100)
    logo = models.ImageField(blank=True, upload_to='django_jwt_oidc')

    def get_logo_as_base64(self, format='png'):
        if self.logo == '' or self.logo is None:
            return ''
        image = self.logo.open(mode='rb')
        encoded_string = base64.b64encode(image.read()).decode('utf-8')
        return 'data:image/%s;base64,%s' % (format, encoded_string)

    def __str__(self):
        return self.host

    def get_name(self):
        if self.name != '':
            return self.name
        return self.host


class PrivateClaimsWebPage(models.Model):
    web = models.ForeignKey(WebPage, on_delete=models.CASCADE)
    attribute_from_user_model = models.CharField(max_length=100, help_text='You can use methods without "()" and use '
                                                                           '"." to get an attribute inside another'
                                                                           ' one.')
    claim = models.CharField(max_length=100, help_text="Claim name. Example: sub, etc.")
    scope = models.CharField(max_length=100, help_text="Extra claim to warn user about what data is being used. "
                                                       "Ex: email, profile, etc.")


class RestrictUsersToWeb(models.Model):
    web = models.ForeignKey(WebPage, on_delete=models.CASCADE)
    attribute_from_user_model = models.CharField(max_length=100, help_text='You can use methods without "()" and use '
                                                                           '"." to get an attribute inside another'
                                                                           ' one.')
    value = models.CharField(max_length=100, help_text='If the user has the same value as this, it won\'t '
                                                       'log in into this web page.')


class UserExternalSession(models.Model):
    id = models.CharField(primary_key=True, default=uuid.uuid4, editable=False, max_length=40)
    access_token_id = models.CharField(default=uuid.uuid4, unique=True, max_length=40)
    authorization_code = models.CharField(blank=True, max_length=40)
    session = models.ForeignKey(Session, on_delete=models.CASCADE)
    web = models.ForeignKey(WebPage, on_delete=models.CASCADE)
    creation_date = models.DateTimeField(default=timezone.now)
    refresh_token = models.IntegerField(default=1)
    code_challenge = models.TextField(blank=True, max_length=100)
    nonce = models.TextField(blank=True, max_length=100)
    id_token_sent = models.BooleanField(default=False)
    access_token_sent = models.BooleanField(default=False)
    scopes = models.TextField(blank=True)

    def get_access_token_id(self):
        self.access_token_id = uuid.uuid4()
        return str(self.access_token_id)

    def get_authorization_code(self):
        self.authorization_code = uuid.uuid4()
        return str(self.authorization_code)

    @classmethod
    def check_authorization_code(cls, code):
        try:
            instance = cls.objects.get(authorization_code=code)
        except cls.DoesNotExist:
            return None
        instance.authorization_code = ''
        instance.save()
        return instance

    def set_code_challenge(self, code_challenge, code_challenge_method):
        self.code_challenge = '%s$%s' % (code_challenge_method, code_challenge)

    def check_code_challenge(self, code_verifier):
        if self.code_challenge == '':
            return False
        alg, digested_hash = self.code_challenge.split('$')
        if alg == 'S256':
            hasher = hashlib.sha256()
            hasher.update(code_verifier.encode('utf-8'))
            return hasher.hexdigest() == digested_hash
        return False

    class Meta:
        unique_together = ('session', 'web')


class WebAllowanceOtherWeb(models.Model):
    web = models.ForeignKey(WebPage, on_delete=models.CASCADE)
    allowed_to = models.ForeignKey(WebPage, on_delete=models.CASCADE, related_name='allowed_on')


class UserWebPagePermission(models.Model):
    web = models.ForeignKey(WebPage, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    scopes_json = models.TextField()

    @classmethod
    def get_not_accepted_and_accepted_scopes(cls, user, web, scopes):
        try:
            instance = cls.objects.get(user=user, web=web)
        except cls.DoesNotExist:
            return scopes, []
        accepted_scopes = instance.scopes
        not_accepted_scopes = [item for item in scopes if item not in accepted_scopes]
        return not_accepted_scopes, accepted_scopes

    @classmethod
    def update_accepted_scopes(cls, user, web, scopes):
        instance = cls.objects.get_or_create(user=user, web=web)[0]
        instance.scopes = scopes
        instance.save()

    @property
    def scopes(self):
        try:
            return json.loads(self.scopes_json)
        except json.JSONDecodeError:
            return []

    @scopes.setter
    def scopes(self, scopes):
        self.scopes_json = json.dumps(scopes)


class NonceUsed(models.Model):
    nonce = models.CharField(max_length=100)
    issued_at = models.DateTimeField(default=timezone.now)

    @classmethod
    def is_used(cls, nonce):
        now = timezone.now()
        key_expiration_time = timezone.timedelta(seconds=get_setting('JWT_OIDC.JWK_EXPIRATION_TIME'))
        max_token_expiration_time = timezone.timedelta(seconds=get_max_time_token() + 120)
        return cls.objects.filter(issued_at__gt=(now - key_expiration_time - max_token_expiration_time),
                                  nonce=nonce).exists()

    @classmethod
    def delete_nonce(cls):
        now = timezone.now()
        key_expiration_time = timezone.timedelta(seconds=get_setting('JWT_OIDC.JWK_EXPIRATION_TIME'))
        max_token_expiration_time = timezone.timedelta(seconds=get_max_time_token() + 120)
        cls.objects.filter(issued_at__lt=(now - key_expiration_time - max_token_expiration_time)).delete()



import hashlib
import json
import logging

import urllib3
from django.core.exceptions import PermissionDenied, BadRequest
from django.http import JsonResponse, HttpResponseBadRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.http import urlencode
from django.views import View
from jwcrypto.jwt import JWTExpired

from django_jwt.auth import JWTAuthentication
from django_jwt.openid import OpenId2Info
from django_jwt.settings_utils import get_setting
from django_jwt.view_utils import crear_url_amb_jwt, get_jwks, get_sub_jwt


logger = logging.getLogger(__name__)


class LoginView(View):
    def get(self, request, *args, **kwargs):
        # id_token, access_token, code parameter only gets here from the OIDC response
        if any([request.GET.get(token, None) is not None for token in ['id_token', 'access_token', 'code']]):
            # User logged in, let's capture the Authentication & Authorization response of the OIDC server
            return self.capture_authentication_and_authorization_response()
        # User did not log in, starting Authentication & Authorization request with the OIDC server
        return self.generate_authentication_and_authorization_request()

    def add_optional_parameter_to_dict(self, params, optional_param):
        optional_param_value = get_setting('JWT_OIDC.CLIENT_' + optional_param.upper())
        if optional_param_value is not None:
            params[optional_param] = optional_param_value

    def generate_authentication_and_authorization_request(self):
        self.request.session['state'] = get_random_string(32)
        self.request.session['nonce'] = get_random_string(64)
        params = {
            'response_type': get_setting('JWT_OIDC.RESPONSE_TYPE'),
            'client_id': get_setting('JWT_OIDC.CLIENT_ID'),
            'redirect_uri': self.request.build_absolute_uri(),
            'scope': get_setting('JWT_OIDC.SCOPE'),
            'state': self.request.session['state'],
            'nonce': self.request.session['nonce'],
        }
        for optional_param in ['display', 'prompt', 'max_age', 'ui_locales', 'claims_locales', 'id_token_hint',
                               'login_hint', 'acr_values']:
            self.add_optional_parameter_to_dict(params, optional_param)
        if get_setting('JWT_OIDC.PKCE_EXTENSION') and 'code' in get_setting('JWT_OIDC.RESPONSE_TYPE').split(' '):
            self.request.session['code_verifier'] = get_random_string(32)
            params['code_challenge_method'] = get_setting('JWT_OIDC.CODE_CHALLENGE_METHOD')
            if params['code_challenge_method'] == 'S256':
                hasher = hashlib.sha256()
                hasher.update(self.request.session['code_verifier'].encode('utf-8'))
                params['code_challenge'] = hasher.hexdigest()
            else:
                logger.error('Invalid CODE_CHALLENGE_METHOD setting: %s.' % params['code_challenge_method'])
                raise Exception('OpenId error')
        return redirect(OpenId2Info().authorization_endpoint + '?' + urlencode(params))

    def capture_and_verify_token(self, token_name, token=None):
        if token is None:
            token = self.request.GET.get(token_name, None)
        try:
            user, jwt = JWTAuthentication.authenticate_credentials(key=token,
                                                                   nonce=self.request.session.get('nonce', ''))
            if user is None:
                raise PermissionDenied()
            changed = False
            for key, value in JWTAuthentication.get_defaults_from_claims(json.loads(jwt.claims))[0].items():
                if getattr(user, key, None) != value:
                    setattr(user, key, value)
                    changed = True
            if changed:
                user.save()
            self.request.session[token_name] = jwt.serialize()
        except (JWTAuthentication.JWTException, JWTExpired):
            logger.error('JWT authentication error.')
            return HttpResponseBadRequest('JWT authentication error.')
        return user

    def capture_authorization_flow_response(self, id_token=None, access_token=None):
        data = {
            'grant_type': 'authorization_code',
            'code': self.request.GET.get('code', None),
            'redirect_uri': self.request.build_absolute_uri(),
            'client_id': get_setting('JWT_OIDC.CLIENT_ID'),
            'client_secret': get_setting('JWT_OIDC.CLIENT_SECRET'),
        }
        if get_setting('JWT_OIDC.PKCE_EXTENSION') and 'code' in get_setting('JWT_OIDC.RESPONSE_TYPE').split(' '):
            data['code_verifier'] = self.request.session['code_verifier']
        headers = {
            'Content-Type': 'application/json',
        }
        http = urllib3.PoolManager()
        r = http.request_encode_body('POST', OpenId2Info().token_endpoint, body=json.dumps(data), headers=headers)
        if r.status != 200:
            raise BadRequest('Unauthorized code')
        response_data = json.loads(r.data.decode('utf-8'))
        if response_data.get('scope', None) and \
                sorted(response_data['scope']) != sorted(get_setting('JWT_OIDC.SCOPE')):
            logger.error('Server returned different scope from configured')
            raise BadRequest('State is not the same!')
        refresh_token = response_data.get('refresh_token', None)
        if refresh_token is not None:
            self.request.session['refresh_token'] = refresh_token
        expires_in = response_data.get('expires_in', None)
        if expires_in is not None:
            self.request.session['expiration_date'] = timezone.now() + timezone.timedelta(seconds=expires_in)
        return response_data.get('access_token', access_token), response_data.get('id_token', id_token)

    def capture_authentication_and_authorization_response(self):
        req_state = self.request.GET.get('state', None)
        if req_state is not None and req_state != self.request.session.get('state', None):
            return HttpResponseBadRequest('State modified!')

        response_type_list = get_setting('JWT_OIDC.RESPONSE_TYPE').split(' ')
        access_token = self.request.GET.get('access_token', None)
        id_token = self.request.GET.get('id_token', None)
        if 'code' in response_type_list:
            access_token, id_token = self.capture_authorization_flow_response(access_token=access_token,
                                                                              id_token=id_token)
        if id_token is not None:
            self.capture_and_verify_token(token_name='id_token', token=id_token)
        if access_token is not None:
            try:
                self.request.session['access_token'] = JWTAuthentication.validate_jwt(access_token).serialize()
            except ValueError:
                self.request.session['access_token'] = access_token
            if 'profile' in get_setting('JWT_OIDC.SCOPE') and OpenId2Info().userinfo_endpoint is not None:
                self.get_userinfo(access_token)
        return redirect(self.request.GET.get('next', '/'))

    def get_userinfo(self, access_token):
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer %s' % access_token
        }
        http = urllib3.PoolManager()
        r = http.request_encode_body('GET', OpenId2Info().userinfo_endpoint, headers=headers)
        if r.status != 200:
            raise BadRequest('Unauthorized code')
        self.request.session['userinfo'] = json.loads(r.data.decode('utf-8'))


class LogoutView(View):
    def get(self, request, *args, **kwargs):
        redirect_uri = get_setting('LOGOUT_REDIRECT_URL')
        if request.user.is_authenticated and OpenId2Info().end_session_endpoint is not None:
            token = request.session.get('id_token', None)
            params = {'post_logout_redirect_uri': request.build_absolute_uri(redirect_uri), 'id_token_hint': token}
            response = redirect(OpenId2Info().end_session_endpoint + '?' + urlencode(params))
        else:
            response = redirect(redirect_uri)
        request.session.flush()
        return response


def fake_config(request):
    config = {
        'issuer': request.build_absolute_uri('/'),
        'userinfo_endpoint': request.build_absolute_uri(reverse('fake_userinfo')),
        'authorization_endpoint': request.build_absolute_uri(reverse('fake_login')),
        'jwks_uri': request.build_absolute_uri(reverse('fake_jwks')),
        'response_types_supported': ['id_token'],
        'subject_types_supported': ['public'],
        'id_token_signing_alg_values_supported': ['RS256'],
        'claims_supported': ['sub', 'iss', 'aud', 'exp', 'iat', 'jti', 'scope', 'azp'],
        'require_request_uri_registration': True,
    }
    return JsonResponse(config)


def fake_login(request):
    if request.method == 'GET':
        return render(request, "django_jwt/fake_login.html")
    if request.method == 'POST':
        return redirect(crear_url_amb_jwt(request))


def jwks(request):
    return JsonResponse(get_jwks())


def fake_userinfo(request):
    sub = get_sub_jwt(request.headers.get('Authorization').split(' ')[1])
    return JsonResponse({'sub': sub})



# Generated by Django 4.0.8 on 2023-04-14 15:07

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('sessions', '0001_initial'),
        ('django_jwt_server', '0002_webpage_logout_all'),
    ]

    operations = [
        migrations.CreateModel(
            name='NonceUsed',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('nonce', models.CharField(max_length=100)),
                ('issued_at', models.DateTimeField(default=django.utils.timezone.now)),
            ],
        ),
        migrations.CreateModel(
            name='PrivateClaimsWebPage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attribute_from_user_model', models.CharField(help_text='You can use methods without "()" and use "." to get an attribute inside another one.', max_length=100)),
                ('claim', models.CharField(help_text='Claim name. Example: sub, etc.', max_length=100)),
                ('scope', models.CharField(help_text='Extra claim to warn user about what data is being used. Ex: email, profile, etc.', max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='RestrictUsersToWeb',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attribute_from_user_model', models.CharField(help_text='You can use methods without "()" and use "." to get an attribute inside another one.', max_length=100)),
                ('value', models.CharField(help_text="If the user has the same value as this, it won't log in into this web page.", max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='WebAllowanceOtherWeb',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.RenameField(
            model_name='userexternalsession',
            old_name='extra_id',
            new_name='access_token_id',
        ),
        migrations.RemoveField(
            model_name='webpage',
            name='needs_confirmation',
        ),
        migrations.RemoveField(
            model_name='webpage',
            name='response_type',
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='access_token_sent',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='authorization_code',
            field=models.CharField(blank=True, max_length=40),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='code_challenge',
            field=models.TextField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='creation_date',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='id_token_sent',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='nonce',
            field=models.TextField(blank=True, max_length=100),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='refresh_token',
            field=models.IntegerField(default=1),
        ),
        migrations.AddField(
            model_name='userexternalsession',
            name='scopes',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='userwebpagepermission',
            name='scopes_json',
            field=models.TextField(default='[]'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='webpage',
            name='client_secret',
            field=models.CharField(default=uuid.uuid4, editable=False, max_length=40),
        ),
        migrations.AddField(
            model_name='webpage',
            name='logo',
            field=models.ImageField(blank=True, upload_to='django_jwt_oidc'),
        ),
        migrations.AddField(
            model_name='webpage',
            name='name',
            field=models.CharField(blank=True, max_length=100),
        ),
        migrations.AlterField(
            model_name='key',
            name='date',
            field=models.DateTimeField(default=django.utils.timezone.now, editable=False),
        ),
        migrations.AlterUniqueTogether(
            name='userexternalsession',
            unique_together={('session', 'web')},
        ),
        migrations.DeleteModel(
            name='AttributeWebPage',
        ),
        migrations.AddField(
            model_name='weballowanceotherweb',
            name='allowed_to',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='allowed_on', to='django_jwt_server.webpage'),
        ),
        migrations.AddField(
            model_name='weballowanceotherweb',
            name='web',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_jwt_server.webpage'),
        ),
        migrations.AddField(
            model_name='restrictuserstoweb',
            name='web',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_jwt_server.webpage'),
        ),
        migrations.AddField(
            model_name='privateclaimswebpage',
            name='web',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_jwt_server.webpage'),
        ),
        migrations.RemoveField(
            model_name='userexternalsession',
            name='date',
        ),
    ]
